{                                                                          }
{ FIXUPBIO  V1.01  Copyright (C) 1997 by Matthias Paul                     }
{                                                                          }
{ A small utility to fixup the (compressed) OpenDOS IBMBIO.COM file to     }
{ enable support for MS Windows95 dual-boot.                               }
{ This code should better be moved into COMPBIOS.EXE, where we also need   }
{ to make some modifications to make it fool-proof...                      }
{                                                                          }
{ FIXUPBIO [<source-file> [<destination-file> [<faked-version>] ] ]        }
{                                                                          }
{ Compile: Borland Pascal 7                                                }
{                                                                          }
{ Last edit: 1997-08-06 -mp                                                }
{                                                                          }
{ Normally, you shouldn't use <faked-version>, since the correct version   }
{ to be faked will be retrieved from IBMBIO.COM itself.  For e.g. 6.00 you }
{ need to enter 1536 or $0600.                                             }

{$A+,B-,D+,E+,F-,G-,I+,L+,N-,O-,P-,Q+,R+,S+,T+,V+,X+}
{$M 16384,150000,655360}

Program FixUpBIO;

 Uses DOS;

 Const
  NUL            = $00;
  LF             = $0D;
  CR             = $0A;
  MainVerStr     = '1';
  SubVerStr      = '01';
  Author         = 'Matthias Paul';
  Description    = 'Utility to fixup Caldera OpenDOS IBMBIO.COM files to enable Win95 dual-boot';
  { Application descriptions for PNW's VERSION.EXE utility }
  NameService    = 'NaMe SeRvIcE='+Description+Chr(NUL)+Chr(CR)+Chr(LF);
  Copyright      = 'CoPyRiGhT=FIXUPBIO (C) 1997 by '+Author+' '+Chr(NUL)+Chr(CR)+Chr(LF);
  Version        = 'VeRsIoN='+MainVerStr+'.'+SubVerStr+' (970806)'+Chr(NUL)+Chr(CR)+Chr(LF);
  PKLITE_Dummy   = 'VeCoP'; { This string is not obsolete!!! }

  compflg      = 3;      { Position of the marker, if the file has been }
                         { compressed                                   }
  JmpInstr     = $E9;
  in_name      : PathStr = 'IBMBIO.COM';
  out_name     : PathStr = 'IBMBIO.COM';

 Type
  file_arr = Array[0..$F000] Of Byte;

 Var
  InFile, OutFile: File;
  buf            : ^file_arr;
  len            : Word;
  value, error   : Word;
  PCDOSsig       : Word;
  io_res         : Integer;
  dummy          : String;

 Function HexPrint(value: LongInt; nr: Byte): String;
  { This one taken from CUI_LIB V1.17.  Returns a hex-string from value. }
  Const
   hex_table: Array [0..15] Of Char = '0123456789ABCDEF';
  Var
   c: Char;
  Begin
   c := hex_table[value AND $0F];
   If (nr > 1) Then HexPrint := Concat(HexPrint(value SHR 4, Pred(nr)), c)
   Else HexPrint := c;
  End; {HexPrint}

 Procedure GetPCDOSSig(Var PCDOSsig: Word);
  Var
   init0 : Word;
  Begin
   Write('  Determining signature...');
   PCDOSsig:=0; { assume error }
   If (buf^[0] = JmpInstr) Then
    Begin
     init0 := buf^[1]+buf^[2]*256+3;
     If ((init0 > 8) And (init0 < len)) Then
      If ((buf^[init0-1] = $0E) And (buf^[init0-2] = $DC) And (buf^[init0-3] >= 2)) Then
       Begin
        PCDOSsig:=buf^[init0-5]*256+buf^[init0-4];
        WriteLn(' ok.');
       End
      Else
       Begin
        WriteLn(' Error!');
        WriteLn('  Signature not found! Probably this is an older IBMBIO.COM file,');
        WriteLn('  not supporting MS Windows95 dual-boot.');
       End
     Else
      WriteLn(' Error! File too short.');
    End
   Else
    WriteLn(' Error! Invalid format of file.');
  End; {GetPCDOSSig}

 Begin {Main}
  { Output banner }
  WriteLn('FIXUPBIO  V', MainVerStr, '.', SubVerStr, '  Copyright (C) 1997 by ', Author);
  WriteLn(' ', Description);

  If (ParamCount >= 1) Then
   Begin
    in_name:=ParamStr(1);
    If (ParamCount >= 2) Then out_name:=ParamStr(2)
    Else out_name:=in_name;
   End;

  { Read in the original compressed file }
  Write('  Reading source file ', in_name, '...');
  {$I-}
  Assign(InFile, in_name);
  Reset(InFile, 1);
  len:=Word(FileSize(InFile));
  GetMem(buf, len);
  BlockRead(InFile, buf^, len);
  {$I+}

  If ((IOResult = 0) And (len > 1024)) Then
   Begin
    {$I-}
    Close(InFile);
    {$I+}
    io_res:=IOResult;

    WriteLn(' ok.');

    error:=1;
    If (ParamCount >= 3) Then
     Begin
      Val(ParamStr(3), value, error);
      If (error = 0) Then PCDOSsig:=value;
     End;
    If (error <> 0) Then
     GetPCDOSSig(PCDOSsig);

    If (PCDOSsig > 0) Then
     Begin
      value:=buf^[compflg]*256+(buf^[Succ(compflg)]);
      If (value = 0) Then
       Begin
        { Insert PC-DOS kernel version signature }
        Write('  Applying patch for PC-DOS real kernel version ',
              HexPrint(Hi(PCDOSsig), 2), '.', HexPrint(Lo(PCDOSsig), 2), '...');
        buf^[compflg]:=Hi(PCDOSsig);
        buf^[Succ(compflg)]:=Lo(PCDOSsig);
        WriteLn(' ok.');

        { Write uncompressed file }
        Write('  Writing destination file ', out_name, '...');

        {$I-}
        Assign(OutFile, out_name);
        Rewrite(Outfile, 1);
        BlockWrite(OutFile, buf^, len);
        {$I+}
        If (IOResult = 0) Then WriteLn(' ok.')
        Else                   WriteLn(' Error!');
        {$I-}
        Close(OutFile);
        {$I+}
        io_res:=IOResult;

       End
      Else
       If (value = (Ord(' ')+256*Ord(' '))) Then
        Begin
         WriteLn('  The source file is copy protected.');
         WriteLn('  Cannot enable dual-boot for MS Windows95.');
        End
       Else
        If (value = PCDOSsig) Then
         Begin
          WriteLn('  The source file already contains fixups for');
          WriteLn('  MS Windows95 dual-boot. Nothing to do.');
         End
        Else
         WriteLn('  The source file has not been compressed. Nothing to do.');
     End
   End
  Else
   Begin
    {$I-}
    Close(InFile);
    {$I+}
    io_res:=IOResult;
    WriteLn(' Error!');
   End;

  FreeMem(buf, len);

  dummy:=PKLITE_Dummy;
  dummy:=NameService;
  dummy:=Version; { for PNW's VERSION.EXE utility }
  dummy:=Copyright;
 End. {Main}
