{                                                                          }
{ UCMPBIOS  V1.01  Copyright (C) 1997 by Matthias Paul                     }
{                                                                          }
{ A crude utility to uncompress the DR DOS 5.0+, Novell DOS, or            }
{ Caldera OpenDOS IBMBIO.COM file, compressed by Andy Wightman's           }
{ compression algorithm.                                                   }
{                                                                          }
{ UCMPBIOS [<source-file> [<destination-file> [<start-offset>] ] ]         }
{                                                                          }
{ Compile: Borland Pascal 7                                                }
{                                                                          }
{ Last edit: 1997-09-19 -mp                                                }
{                                                                          }
{ If not disabled, UCMPBIOS can automatically uncompress IBMBIO.COM files, }
{ that have been prepared for it, that is including the new INFOSTRUC      }
{ (should be all new files after Caldera OpenDOS 7.02).                    }
{ For older files, and files, having been disabled for uncompression,      }
{ you may need to manually change the biosinit_start address, since        }
{ normally the auto-detection leaves too much possible values to try       }
{ (only the first one will be taken).                                      }
{ biosinit_start is the offset in the IBMBIO.COM file, where the           }
{ compression starts and it can be found near by (-0h..-300h) the          }
{ string "$IDLE" in the file, e.g. by comparing files.                     }
{                                                                          }
{ I did not further improve the parameter interface, since this utility    }
{ will most probably be used only once in its life...  The <start_offset>  }
{ will be used only, if UCMPBIOS otherwise would have to guess that value. }

{$A+,B-,D+,E+,F-,G-,I+,L+,N-,O-,P-,Q+,R+,S+,T+,V+,X+}
{$M 16384,150000,655360}

Program UCmpBIOS;

 Uses DOS;

 Const
  NUL            = $00;
  LF             = $0D;
  CR             = $0A;
  MainVerStr     = '1';
  SubVerStr      = '01';
  Author         = 'Matthias Paul';
  Description    = 'Utility to uncompress DR DOS/Novell DOS/Caldera OpenDOS IBMBIO.COM files.';
  { Application descriptions for PNW's VERSION.EXE utility }
  NameService    = 'NaMe SeRvIcE='+Description+Chr(NUL)+Chr(CR)+Chr(LF);
  Copyright      = 'CoPyRiGhT=UCMPBIOS (C) 1997 by '+Author+' '+Chr(NUL)+Chr(CR)+Chr(LF);
  Version        = 'VeRsIoN='+MainVerStr+'.'+SubVerStr+' (970806)'+Chr(NUL)+Chr(CR)+Chr(LF);
  PKLITE_Dummy   = 'VeCoP'; { This string is not obsolete!!! }

  compflg      = 3;      { Position of the marker, if the file has been }
                         { compressed                                   }
  biosinit_start= $5C20; { Starting position of the compressed data     }
                         { $5C20 is for Novell DOS 7 German Update 15/2 }
                         { You may need to change this value, if the    }
                         { auto-detection does not work as desired.     }
  in_name      : PathStr = 'IBMBIO.COM';
  out_name     : PathStr = 'IBMBIO.COM';
  bak_name     : PathStr = 'IBMBIO.BAK';
  JmpInstr     = $E9;

 Type
  file_arr = Array[0..$F000] Of Byte;

 Var
  InFile, OutFile: File;
  in_buf, out_buf: ^file_arr;
  in_len, out_len: Word;
  value, index   : Word;
  biosinit_end   : Word;
  PCDOSsig       : Word;
  io_res         : Integer;
  dummy          : String;

 Function HexPrint(value: LongInt; nr: Byte): String;
  { This one taken from CUI_LIB V1.17.  Returns a hex-string from value. }
  Const
   hex_table: Array [0..15] Of Char = '0123456789ABCDEF';
  Var
   c: Char;
  Begin
   c := hex_table[value AND $0F];
   If (nr > 1) Then HexPrint := Concat(HexPrint(value SHR 4, Pred(nr)), c)
   Else HexPrint := c;
  End; {HexPrint}

 Procedure GetPCDOSSig(Var PCDOSsig: Word);
  { Searches IBMBIO.COM for the new INFOSTRUC and looks up PCDOSsig, }
  { if found. Returns with PCDOSsig=0, if not found.                 }
  Var
   init0 : Word;
  Begin
   PCDOSsig:=0; { assume error }
   If (in_buf^[0] = JmpInstr) Then
    Begin
     init0 := in_buf^[1]+in_buf^[2]*256+3;
     If ((init0 > 8) And (init0 < in_len)) Then
      If ((in_buf^[init0-1] = $0E) And (in_buf^[init0-2] = $DC) And
          (in_buf^[init0-3] >= 2)) Then { structure large enough }
       PCDOSsig:=in_buf^[init0-5]*256+in_buf^[init0-4];
    End;
  End; {GetPCDOSSig}

 Procedure GetBIOSInitEnd(Var biosinit_end: Word);
  { Searches IBMBIO.COM for the new INFOSTRUC, and takes biosinit_end, }
  { if found. Does not change biosinit_end, if not found. Note, that   }
  { the biosinit_end value stored here might be 0, if the file should  }
  { be uncompressable. In this case, we have to guess the correct      }
  { value, e.g. with a little help by UnCompress() with test=TRUE.     }
  Var
   init0 : Word;
  Begin
   If (in_buf^[0] = JmpInstr) Then
    Begin
     init0 := in_buf^[1]+in_buf^[2]*256+3;
     If ((init0 > 8) And (init0 < in_len)) Then
      If ((in_buf^[init0-1] = $0E) And (in_buf^[init0-2] = $DC) And
          (in_buf^[init0-3] >= 4)) Then { structure large enough }
       If ((in_buf^[init0-6] Or in_buf^[init0-7]) > 0) Then
        biosinit_end:=in_buf^[init0-6]*256+in_buf^[init0-7];
    End;
  End; {GetBIOSInitEnd}

 Function Uncompress(biosinit_end: Word; test: Boolean): Word;
  Var
   in_ofs, out_ofs: Word;
   value          : Word;
   ready          : Boolean;
  Begin
   in_ofs :=biosinit_end;
   out_ofs:=biosinit_end;
   If (Not test) Then
    Begin
     Move(in_buf^[0], out_buf^[0], in_ofs); { to play it save, NOP out }
     FillChar(out_buf^[Succ(out_ofs)], SizeOf(file_arr)-out_ofs, $90);
     out_buf^[compflg]      := Lo(biosinit_end); { Reset the flag to the old }
     out_buf^[Succ(compflg)]:= Hi(biosinit_end); { value = start address     }
    End;
   ready:=FALSE;
   If ((in_buf^[in_ofs] Or in_buf^[Succ(in_ofs)]) > 0) Then { skip zeros }
    While ((in_ofs < in_len) And (out_ofs < SizeOf(file_arr)) And (Not ready)) Do
     Begin
      value:=in_buf^[in_ofs];
      Inc(in_ofs);
      value:=value+256*(in_buf^[in_ofs]);
      Inc(in_ofs);
      Case value Of
       $0001..$7FFF: { Copy this many bytes from source to destination }
                    Begin
                     If ((Not test) And ((in_ofs+value) <= in_len)) Then
                      Move(in_buf^[in_ofs], out_buf^[out_ofs], value);
                     Inc(in_ofs, value);
                     Inc(out_ofs, value);
                    End;
       $8000..$FFFF: { Zap in this many zeros }
                    Begin
                     value:= (value AND $7FFF);
                     If ((Not test) And ((out_ofs+value) <= SizeOf(file_arr))) Then
                      FillChar(out_buf^[out_ofs], value, 0);
                     Inc(out_ofs, value);
                    End;
                  0: { Indicates *normal* end }
                    ready:=TRUE;
      End; {Case}
     End;{While}
   If (ready And (in_ofs = in_len-2)) Then Uncompress:=out_ofs
   Else Uncompress:=0; { Indicate error }
  End; {Uncompress}

 Begin {Main}
  { Output banner }
  WriteLn('UCMPBIOS  V', MainVerStr, '.', SubVerStr, '  Copyright (C) 1997 by ', Author);
  WriteLn(' ', Description);

  If (ParamCount >= 1) Then
   Begin
    in_name:=ParamStr(1);
    If (ParamCount >= 2) Then out_name:=ParamStr(2)
    Else out_name:=in_name;
   End;

  { Read in the original compressed file }
  Write('  Reading source file ', in_name, '...');
  {$I-}
  Assign(InFile, in_name);
  Reset(InFile, 1);
  in_len:=Word(FileSize(InFile));
  GetMem(in_buf, in_len);
  BlockRead(InFile, in_buf^, in_len);
  {$I+}

  If (IOResult = 0) Then
   Begin
    {$I-}
    Close(InFile);
    {$I+}
    io_res:=IOResult;

    WriteLn(' ok.');
    value:=in_buf^[compflg]*256+(in_buf^[Succ(compflg)]);
    GetPCDOSSig(PCDOSsig);
    If (((value = 0) Or (value = PCDOSsig)) And (value <> (Ord(' ')+256*Ord(' '))) And
       (in_buf^[in_len-2] = 0) And (in_buf^[Pred(in_len)] = 0)) Then
     Begin
      biosinit_end:=0; { assume uncompression was disabled for the file }
      If (ParamCount >= 3) Then
       Begin
        Val(ParamStr(3), value, index);
        If (index = 0) Then biosinit_end:=value;
       End;
      GetBIOSInitEnd(biosinit_end);
      If (biosinit_end = 0) Then
       Begin
        Write('  Guessing <biosinit_end> offset, trying only first possible match...');
        biosinit_end:=biosinit_start;
        While ((biosinit_end < in_len) And (UnCompress(biosinit_end, TRUE) = 0)) Do
         Inc(biosinit_end); { Only the first possible offset will be used,   }
                            { which must not necessarily be the correct one! }
                            { Unfortunately, we have to guess a little here. }
        WriteLn(' done.');
        WriteLn('  Warning: Uncompression based on guessed offsets most probably');
        WriteLn('           will result in invalid files!');
        If (ParamCount <= 1) Then
         Begin
          Write('  Writing backup file ', bak_name, '...');
          {$I-}
          Assign(OutFile, bak_name);
          Rewrite(Outfile, 1);
          BlockWrite(OutFile, in_buf^, in_len);
          {$I+}
          If (IOResult = 0) Then WriteLn(' ok.')
          Else WriteLn(' Error!');
          {$I-}
          Close(OutFile);
          {$I+}
          io_res:=IOResult;
         End;
       End
      Else
       WriteLn('  Retrieving <biosinit_end> offset... ok.');

      Write('  Checking...');
      If ((biosinit_end >= biosinit_start) And (biosinit_end < in_len)) Then
       Begin
        WriteLn(' ok.');
        Write('  Uncompressing file from address ', HexPrint(biosinit_end, 4), 'h onwards...');
        { Prepare for uncompression }
        GetMem(out_buf, SizeOf(file_arr));

        out_len:=UnCompress(biosinit_end, FALSE);

        If (out_len > 0) Then WriteLn(' ok.')
        Else WriteLn(' Error! Invalid format of source file.');

        { Write uncompressed file }
        Write('  Writing destination file ', out_name, '...');
        {$I-}
        Assign(OutFile, out_name);
        Rewrite(Outfile, 1);
        BlockWrite(OutFile, out_buf^, out_len);
        {$I+}
        If (IOResult = 0) Then WriteLn(' ok.')
        Else                   WriteLn(' Error!');
        {$I-}
        Close(OutFile);
        {$I+}
        io_res:=IOResult;

        FreeMem(out_buf, SizeOf(file_arr));
       End
      Else WriteLn(' Error! Invalid format of source file.');
     End
    Else
     If (value = (Ord(' ')+256*Ord(' '))) Then
      WriteLn('  The source file is copy protected. Cannot uncompress.')
     Else
      WriteLn('  The source file has not been compressed. Nothing to do.');
   End
  Else
   Begin
    {$I-}
    Close(InFile);
    {$I+}
    io_res:=IOResult;
    WriteLn(' Error!');
   End;

  FreeMem(in_buf, in_len);

  dummy:=PKLITE_Dummy;
  dummy:=NameService;
  dummy:=Version; { for PNW's VERSION.EXE utility }
  dummy:=Copyright;

 End. {Main}
